//  bbdbits.cpp - ( part of class cDemuxer, bbdmux12.h, bbdmux12.cpp
//
// MPEG2AVI
// --------
// 
// v0.16B34
//   modified bs_getbytes() to allow fast skipping of bytes
//   if the parameter (*pcbStored ==NULL), skip N bytes
//   converted malloc() calls to calloc() calls
//   optimized bs_fseek_lba() for special case, bs_update_lba() modified
//   when the (seek-target LBA) == (current LBA + 1), bs_fseek_lba uses
//   bs_getbytes to "skip" to the next LBA, rather then using the slower
//	   my_seeki64() call.
//   
// v0.16B32
//   replaced the _inline functions of my_open, my_read, my_close with
//   #defines, to reduce calling overhead
// v0.16B3
//   added logging via TextMsgLog(), TextMsgErr() to most functions
//   did not add these calls to speed critical functions
//   fixed bs_seek_lba()
//   improved bs_getbytes(), manually orders bs_refill_buffer() if the
//    buffer is too small
// v0.16B24 converted bs_ package to use _open, _read, _lseek, and _close
//   instead of the higher-level fopen, fread, etc.
//   Changed bs_refill() to use a *single* _read() of N bytes,
//    rather than N _read()'s of 1-byte each.  bs_refill() now returns a value
//    lba_fseek() now uses _lseeki64(), but not tested.  This should allow
//    seeking through bitstreams >4GB length
// v0.16B23
//   inlined bs_getbits()
// v0.16B22
//   added bs_getbytes(), for faster reading of long packets (>1 bytes)
//    bs_getbytes() checks for bitstream byte-alignment, and if this
//    is true, bs_getbytes() uses its own loop to extract bytes directly
//    from the bs->buf[].
//    Otherwise, bs_getbytes() calls bs_getbits( 8 ) repeatedly (in
//    which case the performance is much worse.)
// v0.16B2
//   converted to C++,
//   all functions start with the prefix "bs_...()", to help distinguish
//   the bitstream-related functions with the demuxer-functions
//   updated s_fwrite() to write into a local buffer (supplie dby go_demux()),
//   no more direct manipulation of extern g_demuxbuffer()!
//
//   There are three files for class cDemuxer : 
//    bbdmux12.cpp, bbdmux12.h, bbdbits.cpp
//
// v0.13 initial release
//  (originally from Brent Beyeler's BBDMUX v1.1 MPEG-2 demuxer package)
//
//  all fwrite() calls have been replaced with "s_fwrite()"
//  s_fwrite() stores data in the demux_buffer(), instead of writing to a
//  file.  Hence, s_fwrite() calls demux_buffer::push_data()
//

/*****************************************************************************
*
*  bit_stream.c package
*  Author:  Jean-Georges Fritsch, C-Cube Microsystems
*
*****************************************************************************/

/********************************************************************
  This package provides functions to write (exclusive or read)
  information from (exclusive or to) the bit stream.

  If the bit stream is opened in read mode only the get functions are
  available. If the bit stream is opened in write mode only the put
  functions are available.
********************************************************************/

/*open_bit_stream_w(); open the device to write the bit stream into it    */
/*bs_open_bit_stream_r(); open the device to read the bit stream from it     */
/*bs_close_bit_stream();  close the device containing the bit stream         */
/*bs_alloc_buffer();      open and initialize the buffer;                    */
/*bs_desalloc_buffer();   empty and close the buffer                         */
/*bs_back_track_buffer();     goes back N bits in the buffer                 */
/*unsigned int bs_get1bit();  read 1 bit from the bit stream                 */
/*unsigned long bs_getbits(); read N bits from the bit stream                */
/*unsigned long bs_byte_ali_getbits();   read the next byte aligned N bits from*/
/*                                    the bit stream                        */
/*unsigned long bs_look_ahead(); grep the next N bits in the bit stream without*/
/*                            changing the buffer pointer                   */
/*put1bit(); write 1 bit from the bit stream  */
/*put1bit(); write 1 bit from the bit stream  */
/*putbits(); write N bits from the bit stream */
/*byte_ali_putbits(); write byte aligned the next N bits into the bit stream*/
/*unsigned long bs_sstell(); return the current bit stream length (in bits)    */
/*int bs_end_bs(); return 1 if the end of bit stream reached otherwise 0       */
/*int bs_seek_sync(); return 1 if a sync word was found in the bit stream      */
/*                 otherwise returns 0                                      */


//put1bit()
//putbits()
//byte_ali_putbits()

#include "bbdmux12.h"

#include <io.h>
#include <fcntl.h>
#include <stdio.h>

#define min(a, b)  (((a) < (b)) ? (a) : (b)) 

	// my_open(), my_close(), and my_read() are located in mpv_read.cpp
	// the cDemuxer doesn't directly call the I/O routines, because 
	// the calls are instead routed through the cStreamlist object.
	// The streamlist object effectively concatenate a group of files into
	// one "logical file"
	extern "C" int my_open( const char *filename, int oflag );
	extern "C" int my_close( int handle );
	extern "C" int my_read( int handle, void *buffer, unsigned int count );
	extern "C" __int64 my_lseeki64( int handle, __int64 offset, int origin );

	// uncomment these for link-up with the streamlist junk in mpv_read.cpp
	#define _b_open(filename,oflag) (my_open(filename,oflag))
	#define _b_close(handle) (my_close(handle))
	#define _b_read(handle,buffer,count) (my_read(handle,buffer,count))

	// uncomment these lines to direct file-access
	//#define _b_open(filename,oflag) (_open(filename,oflag))
	//#define _b_close(handle) (_close(handle))
	//#define _b_read(handle,buffer,count) (_read(handle,buffer,count);


// s_fwrite() replaces all instances of fwrite() in the original source
// code.  Originally, s_fwrite() called demux_buffer::push_data(), storing
// the MPEG-2 (demuxed) stream data into our demux_buffer (instead of
// writing a file.)
//
// Starting with v0.16B2, s_fwrite() pushes data into an internal temp
// storage space (set by go_demux().)  

// < CURRENT IMPLEMENTATION AT END OF FILE >

//size_t
//s_fwrite( const void *buffer, size_t size, size_t count, FILE *stream );
//{
	//return fwrite( buffer, size, count, stream );
//}



/*******************************************************************************
*
*  Allocate number of bytes of memory equal to "block".
*
*******************************************************************************/

// returns NULL if failure
void *cDemuxer::bs_mem_alloc(
unsigned long block,
char *item)
{
  void *ptr;

  sprintf( m_szText, "bs_mem_alloc( %u ) : ", block );
  TextLogMsg( LOG_NEW_ENTRY, m_szText );

  ptr = (void *) calloc((unsigned int)(block>>5)+1, 32 );
  if (ptr != NULL)
    memset(ptr, 0, (unsigned int)block);
  else
  {
    sprintf( m_szText, "bs_mem_alloc() : malloc(%u) failed", block);
	TextLogErr( BSERR_FAIL, m_szText );
	return NULL;
  }

  sprintf( m_szText, "allocated %u bytes.", block );
  TextLogMsg( 0, m_szText );

  return(ptr);
}


/****************************************************************************
*
*  Free memory pointed to by "*ptr_addr".
*
*****************************************************************************/

void cDemuxer::bs_mem_free(
void **ptr_addr)
{
  TextLogMsg( LOG_NEW_ENTRY, "bs_mem_free() : ");

  if (*ptr_addr != NULL)
  {
    free(*ptr_addr);
    *ptr_addr = NULL;
	TextLogMsg( 0, "deallocated *ptr_addr[]");
  }
  else
	  TextLogErr( BSERR_FAIL, "bs_mem_free() : ptr_addr == NULL!");
}

/*******************************************************************************
*
*  Check block of memory all equal to a single byte, else return false
*
*******************************************************************************/

int cDemuxer::bs_memcheck(
char *array,
int test,       /* but only tested as a char (bottom 8 bits) */
int num)
{
  int i = 0;
  sprintf( m_szText, "bs_memcheck( %d, %d ) : ", test, num );
  TextLogMsg( LOG_NEW_ENTRY, m_szText );

  while (array[i] == test && i<num)
    i++;
  if (i == num)
    return true;
  else
    return false;
}



/* refill the buffer from the input device when the buffer becomes empty    */
DWORD cDemuxer::bs_refill_buffer(
Bit_stream_struc *pbs)   /* bit stream structure */
{
  int i = pbs->buf_size - 2 - pbs->buf_byte_idx;
  int n = 0;
  int index = 0;
  char val[2];
  unsigned char inbuf[ DEMUX_BUFFER_SIZE ];

  if ( !BSOPEN( pbs ) )
  {
	  TextLogErr( BSERR_FAIL, "bs_refill_buffer() : no bitstream is open." );
	  return BSERR_FAIL;
  }

  if (pbs->format != DEMUX_BINARY)
  {
	  TextLogErr( BSERR_FAIL, "bs_refill_buffer() : bitstream not in BINARY mode!" );
	  return BSERR_FAIL;
  }

//  printf("\nbuf_size = %d, buf_byte_idx = %d", pbs->buf_size ,
//		pbs->buf_byte_idx );
//  printf("\ncalling _read( %d, inbuf, %d )", pbs->fhandle, i+1 );
  n = _b_read( pbs->fhandle, inbuf, (i+1) );
//  printf(" ... got %d", n );

  if ( n == -1 ) // check if function even succeeded? 
  {		//file handle not valid!
	  TextLogErr( BSERR_FAIL, "bs_refill_buffer() : File handle not valid!" );
	  return BSERR_FAIL;
  }

  if ( n != (i+1) )
  {
	fprintf( stderr, "\nbs_refill_buffer() : EOF !");
	// this is not an end-of-bitstream, because the bitstream position has
	// not yet reached the end of the file (even though the read-ahead has
	// reached EOF
	//pbs->eobs = 1; // signify end of bitstream, v0.16B3
	pbs->eob = i - (n);
  }

  for ( index = 0; index < n; ++index )
  {
	  pbs->buf[ i ] = inbuf[ index ];
	 --i;
  }	// copy buffer

/*
  while ((i >= 0) && (!pbs->eob))
  {
    if (pbs->format == DEMUX_BINARY)
      //n = fread(&pbs->buf[i--], sizeof(unsigned char), 1, pbs->pt);
	  n = _read( pbs->fhandle, &pbs->buf[i--], 1 );
    else
    {
      while((index < 2) && n)
      {
        //n = fread(&val[index], sizeof(char), 1, pbs->pt);
		n = _read( pbs->fhandle, &val[index], 1 );
        switch (val[index])
        {
          case 0x30:
          case 0x31:
          case 0x32:
          case 0x33:
          case 0x34:
          case 0x35:
          case 0x36:
          case 0x37:
          case 0x38:
          case 0x39:
          case 0x41:
          case 0x42:
          case 0x43:
          case 0x44:
          case 0x45:
          case 0x46:
            index++;
            break;
          default:
            break;
        } // endswitch
      }  // end while((index < 2) && n)

      if (val[0] <= 0x39)
        pbs->buf[i] = (val[0] - 0x30) << 4;
      else
        pbs->buf[i] = (val[0] - 0x37) << 4;
      if (val[1] <= 0x39)
        pbs->buf[i--] |= (val[1] - 0x30);
      else
        pbs->buf[i--] |= (val[1] - 0x37);
      index = 0;
    } // end if (bs->format == DEMUX_BINARY)


    if (!n) // were we succesfull in reading the byte?
	{
      pbs->eob = i + 1;
	  pbs->eobs = 1;	// signify end of bitstream
	}
  }*/
  return BSERR_OK;
}


/* resets input buffer state to empty */
void 
cDemuxer::bs_reset_buffer(
Bit_stream_struc *pbs)   /* bit stream structure */
{
  TextLogMsg( LOG_NEW_ENTRY, "bs_reset_buffer() : " );
  pbs->buf_byte_idx = 0;
  pbs->buf_bit_idx = 0;
  pbs->eob = false;
  pbs->eobs = false;
}


/* empty the buffer to the output device when the buffer becomes full */
//void empty_buffer(
//Bit_stream_struc *bs,   /* bit stream structure */
//int minimum)            /* end of the buffer to empty */
//{
//  register int i;
//
//#if BS_FORMAT == BINARY
//  for (i = bs->buf_size - 1; i >= minimum; i--)
//    fwrite(&bs->buf[i], sizeof(unsigned char), 1, bs->pt);
//#else
//  for (i = bs->buf_size - 1; i >= minimum; i--)
//  {
//    char val[2];
//    val[0] = he[((bs->buf[i] >> 4) & 0x0F)];
//    val[1] = he[(bs->buf[i] & 0x0F)];
//    fwrite(val, sizeof(char), 2, bs->pt);
//  }
//#endif
//
//  for (i = minimum - 1; i >= 0; i--)
//    bs->buf[bs->buf_size - minimum + i] = bs->buf[i];

//  bs->buf_byte_idx = bs->buf_size -1 - minimum;
//  bs->buf_bit_idx = 8;
//}

/* open the device to write the bit stream into it */
//void open_bit_stream_w(
//Bit_stream_struc *bs,   /* bit stream structure */
//const char *bs_filenam,       /* name of the bit stream file */
//int size)               /* size of the buffer */
//{
//  if ((bs->pt = fopen(bs_filenam, "wb")) == NULL)
//  {
//    printf("Could not create \"%s\".\n", bs_filenam);
//    exit(1);
//  }
//  bs_alloc_buffer(bs, size);
//  bs->buf_byte_idx = size-1;
//  bs->buf_bit_idx = 8;
//  bs->totbit = 0;
//  bs->mode = WRITE_MODE;
//  bs->eob = false;
//  bs->eobs = false;
//}

/* open the device to read the bit stream from it */
long cDemuxer::bs_open_bit_stream_r(
Bit_stream_struc *pbs,   /* bit stream structure */
//const char *bs_filenam,       /* name of the bit stream file */
int size)               /* size of the buffer */
{
  register unsigned long n;
  register unsigned char flag = 1;
  unsigned char val;

  TextLogMsg( LOG_NEW_ENTRY, "bs_open_bit_stream_r() : " );

  if ( BSOPEN( pbs ) )
  {
	TextLogErr( BSERR_FILE_OPEN, "bs_open_bit_stream_r() : Another bitstream is already open!");
	return BSERR_FILE_OPEN;
  }

  pbs->fhandle = _b_open( pbs->filename, _O_BINARY | _O_RDONLY );
  //bs->pt = fopen(bs->filename, "rb" );
  if ( !BSOPEN( pbs ) )
  {
    sprintf( m_szText, "bs_open_bit_stream_r() : Could not open \"%s\".\n",
		pbs->filename );
	TextLogErr( BSERR_FILE_NOT_FOUND, m_szText );
    return BSERR_FILE_NOT_FOUND;
  }

  // close and reopen file...
  // mpeg2dec.c may have already opened the file, thus we need
  // to close it and reopen it to reset the file-pointer
  //fclose ( bs->pt );
  //bs->pt = fopen(bs->filename, "rb");

  do
  {
   // n = fread(&val, sizeof(unsigned char), 1, bs->pt);
	n = _read( pbs->fhandle, &val, 1 );
    switch (val)
    {
      case 0x30:
      case 0x31:
      case 0x32:
      case 0x33:
      case 0x34:
      case 0x35:
      case 0x36:
      case 0x37:
      case 0x38:
      case 0x39:
      case 0x41:
      case 0x42:
      case 0x43:
      case 0x44:
      case 0x45:
      case 0x46:
      case 0xa:  /* \n */
      case 0xd:  /* cr */
      case 0x1a:  /* sub */
        break;

      default: /* detection of an binary character */
        flag--;
        break;
    }
  } while (flag & n);

  if (flag)
    pbs->format = DEMUX_ASCII;
  else
    pbs->format = DEMUX_BINARY;

  // now derive the bitstream's length by seeking to its end
  pbs->qwLength = my_lseeki64( pbs->fhandle, 0, SEEK_END );
  if ( pbs->qwLength == -1 )
  {
	TextLogErr( BSERR_FAIL, "bs_open_bit_stream_r() : _lseeki64() failed, bitstream length UNKNOWN." );
	//bs_close_bit_stream_r( pbs);
	return BSERR_FAIL;
  }

  bs_close_bit_stream_r( pbs);

  // reopen the bitstream
  pbs->fhandle = _b_open( pbs->filename, _O_BINARY | _O_RDONLY );
  if ( !BSOPEN( pbs ) )
  {
	  
    sprintf( m_szText, "bs_open_bit_stream_r() : Could not open \"%s\".\n", 
		pbs->filename);
	TextLogErr( BSERR_FILE_NOT_FOUND, m_szText );
    return BSERR_FILE_NOT_FOUND;
  }

  bs_alloc_buffer(pbs, size);
  bs_reset_buffer( pbs );
  pbs->mode = DEMUX_READ_MODE;
  pbs->totbit = 0;

  return BSERR_OK;
}

/*close the device containing the bit stream after a read process*/
long cDemuxer::bs_close_bit_stream_r(
Bit_stream_struc *pbs)   /* bit stream structure */
{
  sprintf( m_szText, "bs_close_bit_stream_r( '%s' ) : ", pbs->filename );
  TextLogMsg( LOG_NEW_ENTRY, m_szText );
  if ( !BSOPEN( pbs ) )
  {
	  TextLogErr( BSERR_FAIL, "bs_close_bit_stream_r() : no bitstream was open!" );
	  return BSERR_FAIL;
  }
  _b_close( pbs->fhandle );
  pbs->fhandle = _NOFILE;	// clear file handle
//  fclose(bs->pt);
  bs_desalloc_buffer(pbs);


  //bs->pt = NULL;	// delete the FILE* pointer (other functions use this 
					// condition to signal that no bitstream is open.

  return BSERR_OK;
}

/*close the device containing the bit stream after a write process*/
//void close_bit_stream_w(
//Bit_stream_struc *bs)   /* bit stream structure */
//{
//  empty_buffer(bs, bs->buf_byte_idx);
//  fclose(bs->pt);
//  bs_desalloc_buffer(bs);
//}

/*open and initialize the buffer; */
void cDemuxer::bs_alloc_buffer(
Bit_stream_struc *pbs,   /* bit stream structure */
int size)
{
  sprintf( m_szText, "bs_alloc_buffer( %d ) : ", size );
  TextLogMsg( LOG_NEW_ENTRY, m_szText );
  if ( pbs-> buf != NULL )
  {
	TextLogMsg( 0, "Buffer already allocated!  De-allocating & re-allocating..." );
	bs_desalloc_buffer( pbs );
  }

  pbs->buf = (unsigned char *) bs_mem_alloc(size * sizeof(unsigned char), "buffer");

  pbs->buf_size = size;
}

/*empty and close the buffer */
void cDemuxer::bs_desalloc_buffer(
Bit_stream_struc *pbs)   /* bit stream structure */
{
  TextLogMsg( LOG_NEW_ENTRY, "bs_desalloc_buffer() : " );

  if ( pbs->buf != NULL )
  {
	  free(pbs->buf);
	  TextLogMsg( 0, "deallocated buffer." );
  }
  else
	  TextLogMsg( 0, "nothing to deallocate!" );
  pbs->buf = NULL;
}

int bs_putmask[9] = {0x0, 0x1, 0x3, 0x7, 0xf, 0x1f, 0x3f, 0x7f, 0xff};
int demux_clearmask[9] = {0xff, 0xfe, 0xfc, 0xf8, 0xf0, 0xe0, 0xc0, 0x80, 0x0};

//void bs_back_track_buffer( /* goes back N bits in the buffer */
//Bit_stream_struc *bs,   /* bit stream structure */
//int N)
//{
//  int tmp = N - (N / 8) * 8;
  //register int i;

  //bs->totbit -= N;
  //for (i = bs->buf_byte_idx; i < bs->buf_byte_idx + N / 8 - 1; i++)
//    bs->buf[i] = 0;
//  bs->buf_byte_idx += N / 8;
//  if ((tmp + bs->buf_bit_idx) <= 8)
    //bs->buf_bit_idx += tmp;
  //else
  //{
//    bs->buf_byte_idx ++;
    //bs->buf_bit_idx += (tmp - 8);
  //}
  //bs->buf[bs->buf_byte_idx] &= demux_clearmask[bs->buf_bit_idx];
//}

unsigned int demux_mask[8]={0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x80};

/*read 1 bit from the bit stream */
unsigned int cDemuxer::bs_get1bit(
Bit_stream_struc *bs)   /* bit stream structure */
{
  unsigned int bit;
  register int i;

  bs->totbit++;
  if (!bs->buf_bit_idx)
  {
    bs->buf_bit_idx = 8;
    bs->buf_byte_idx--;
	++bs->dwBytePos;
	bs_update_lba( bs );
    if ((bs->buf_byte_idx < DEMUX_MINIMUM) || (bs->buf_byte_idx < bs->eob))
    {
      if (bs->eob)
        bs->eobs = true;
      else
      {
        for (i = bs->buf_byte_idx; i >= 0; i--)
          bs->buf[bs->buf_size - 1 - bs->buf_byte_idx+i] = bs->buf[i];
        bs_refill_buffer(bs);
        bs->buf_byte_idx = bs->buf_size-1;
      }
    }
  }
  bit = bs->buf[bs->buf_byte_idx] & demux_mask[bs->buf_bit_idx - 1];
  bit = bit >> (bs->buf_bit_idx - 1);
  bs->buf_bit_idx--;
  return(bit);
}

/*write 1 bit from the bit stream */
//void put1bit(
//Bit_stream_struc *bs,   /* bit stream structure */
//int bit)                /* bit to write into the buffer */
//{f
//  bs->totbit++;
//  bs->buf[bs->buf_byte_idx] |= (bit & 0x1) << (bs->buf_bit_idx - 1);
//  bs->buf_bit_idx--;
//  if (!bs->buf_bit_idx)
//  {
//    bs->buf_bit_idx = 8;
//    bs->buf_byte_idx--;
//    if (bs->buf_byte_idx < 0)
//      empty_buffer(bs, MINIMUM);
//    bs->buf[bs->buf_byte_idx] = 0;
//  }
//}

/*look ahead for the next N bits from the bit stream */
unsigned long cDemuxer::bs_look_ahead(
Bit_stream_struc *bs,   /* bit stream structure */
int N)                  /* number of bits to read from the bit stream */
{
  unsigned long val = 0;
  register int j = N;
  register int k, tmp;
  register int bit_idx = bs->buf_bit_idx;
  register int byte_idx = bs->buf_byte_idx;

  if (N > DEMUX_MAX_LENGTH)
  {
    sprintf( m_szText, "bs_look_ahead( %d ) : Cannot read or write more than %d bits at a time.", 
		N, DEMUX_MAX_LENGTH);
	TextLogErr( BSERR_FAIL, m_szText );
    exit(1);
  }

  while (j > 0)
  {
    if (!bit_idx)
    {
      bit_idx = 8;
      byte_idx--;
    }
    k = min(j, bit_idx);
    tmp = bs->buf[byte_idx] & bs_putmask[bit_idx];
    tmp = tmp >> (bit_idx - k);
    val |= tmp << (j - k);
    bit_idx -= k;
    j -= k;
  }
  return(val);
}

/*read N bit from the bit stream */
unsigned long cDemuxer::bs_getbits(
Bit_stream_struc *bs,   /* bit stream structure */
int N)                  /* number of bits to read from the bit stream */
{
  unsigned long val = 0;
  register int i;
  register int j = N;
  register int k, tmp;

  if (N > DEMUX_MAX_LENGTH)
  {
    sprintf( m_szText, "bs_getbits( %d ) : Cannot read or write more than %d bits at a time.", 
		N, DEMUX_MAX_LENGTH);
	TextLogErr( BSERR_FAIL, m_szText );
    exit(1);
  }
  bs->totbit += N;
  while (j > 0)
  {
    if (!bs->buf_bit_idx)
    {
      bs->buf_bit_idx = 8;
      bs->buf_byte_idx--;
	  ++bs->dwBytePos;
//	  bs_update_lba( bs );
      if ((bs->buf_byte_idx < DEMUX_MINIMUM) || (bs->buf_byte_idx < bs->eob))
      {
        if (bs->eob)
          bs->eobs = true;
        else
        {
          for (i = bs->buf_byte_idx; i >= 0; i--)
            bs->buf[bs->buf_size - 1 - bs->buf_byte_idx+i] = bs->buf[i];
          bs_refill_buffer(bs);
          bs->buf_byte_idx = bs->buf_size - 1;
        }
      }
    }
    k = min(j, bs->buf_bit_idx);
    tmp = bs->buf[bs->buf_byte_idx] & bs_putmask[bs->buf_bit_idx];
    tmp = tmp >> (bs->buf_bit_idx - k);
    val |= tmp << (j - k);
    bs->buf_bit_idx -= k;
    j -= k;
  }

  bs_update_lba( bs );

  return(val);
}

/*write N bits into the bit stream */
//void putbits(
//Bit_stream_struc *bs,   /* bit stream structure */
//unsigned int val,       /* val to write into the buffer */
//int N)                  /* number of bits of val */
//{
//  register int j = N;
//  register int k, tmp;
//
//  if (N > MAX_LENGTH)
//  {
    //printf("Cannot read or write more than %d bits at a time.\n", MAX_LENGTH);
  //exit(1);
 //}
//
//  bs->totbit += N;
//  while (j > 0)
//  {
//    k = min(j, bs->buf_bit_idx);
//   tmp = val >> (j - k);
//    bs->buf[bs->buf_byte_idx] |= (tmp & bs_putmask[k]) << (bs->buf_bit_idx - k);
//    bs->buf_bit_idx -= k;
//    if (!bs->buf_bit_idx)
//    {
//      bs->buf_bit_idx = 8;
//      bs->buf_byte_idx--;
//      if (bs->buf_byte_idx < 0)
//        empty_buffer(bs, MINIMUM);
//      bs->buf[bs->buf_byte_idx] = 0;
//    }
//    j -= k;
// }
//}

/* do a mod on a floating point */
unsigned long 
cDemuxer::bs_mod_double(double val, double modval)
{
  return (unsigned long) floor(val - floor(val / modval) * modval);
}


/*write N bits byte aligned into the bit stream */
//void byte_ali_putbits(
//Bit_stream_struc *bs,   /* bit stream structure */
//unsigned int val,       /* val to write into the buffer */
//int N)                  /* number of bits of val */
//{
//  unsigned long aligning;

//  if (N > MAX_LENGTH)
//  {
//    printf("Cannot read or write more than %d bits at a time.\n", MAX_LENGTH);
//    exit(1);
//  }

//  aligning = bs_mod_double(bs_sstell(bs), 8.0);
//  if (aligning)
//    putbits(bs, (unsigned int)0, (int)(8 - aligning));

//  putbits(bs, val, N);
//}

/*read the next bute aligned N bits from the bit stream */
unsigned long cDemuxer::bs_byte_ali_getbits(
Bit_stream_struc *bs,   /* bit stream structure */
int N)                  /* number of bits of val */
{
  unsigned long aligning;

  if (N > DEMUX_MAX_LENGTH)
  {
    printf("Cannot read or write more than %d bits at a time.\n", 
		DEMUX_MAX_LENGTH);
    exit(1);
  }

  aligning = bs_mod_double(bs_sstell(bs), 8.0);
  if (aligning)
    bs_getbits(bs, (int)(8 - aligning));

  return (bs_getbits(bs, N));
}

/*return the current bit stream length (in bits)*/
double cDemuxer::bs_sstell(
Bit_stream_struc *bs)   /* bit stream structure */
{
  return(bs->totbit);
}

/*return the status of the bit stream*/
/* returns 1 if end of bit stream was reached */
/* returns 0 if end of bit stream was not reached */
int cDemuxer::bs_end_bs(
Bit_stream_struc *bs)   /* bit stream structure */
{
  return(bs->eobs);
}

// moved to header file
//static unsigned int val;

/*this function seeks for a byte aligned sync word in the bit stream and
  places the bit stream pointer right after the sync.
  This function returns 1 if the sync was found otherwise it returns 0  */
int cDemuxer::bs_seek_sync(
Bit_stream_struc *bs,   /* bit stream structure */
unsigned int sync,      /* sync word maximum 32 bits */
int N)          /* sync word length */
{
  unsigned int aligning;
  unsigned int val;
  unsigned int maxi = (int) pow(2.0, (double) N) - 1;

  sprintf( m_szText, "bs_seek_sync( sync_word(0x%X), len=%d bits ) : ",
	  sync, N );
  TextLogMsg( LOG_NEW_ENTRY, m_szText );

  aligning = bs_mod_double(bs_sstell(bs), (double) DEMUX_ALIGNING);
  if (aligning)
    bs_getbits(bs, (int)(DEMUX_ALIGNING - aligning));

  val = bs_getbits(bs, N);
  while (((val & maxi) != sync) && (!bs_end_bs(bs)))
  {
    val <<= DEMUX_ALIGNING;
    val |= bs_getbits(bs, DEMUX_ALIGNING);
  }

  bs_update_lba( bs );	// update position
  if (bs_end_bs(bs))
    return(0);
  else
    return(1);
}

/*****************************************************************************
*
*  End of bit_stream.c package
*
*****************************************************************************/


/*  this version of s_fwrite() emulates the *fwrite()* command 
size_t
cDemuxer::s_fwrite( const void *buffer, size_t size, size_t count, FILE *stream )
{
	return fwrite( buffer, size, count, stream );
}
*/


// normally, s_fwrite() is called by go_demux() and its subfunctions.
// FwriteBuf is only valid for the duration of go_demux()'s invocation.
// after go_demux() terminates, FwriteBuf becomes invalid.
size_t
cDemuxer::s_fwrite( unsigned char *buffer, size_t size, size_t count, FILE *stream )
{
	unsigned long rc = 0;
	unsigned long i;
	unsigned char *lpchar = buffer;

	if ( buffer == NULL || m_pFwriteBuf == NULL )
	{
		TextLogErr( BSERR_FAIL, "cDemuxer::s_fwrite() : NULL pointers!");
		return 0;	// failed to write anything, return 0!
	}

	// copy input data (buffer[]) to Fwritebuf[]
	for( i = 0; i < size*count; ++i )
	{
		m_pFwriteBuf[ i + m_cbFwriteBuf ] = lpchar[ i ] ;  
	}

	m_cbFwriteBuf += i;	// update the #bytes stored in Fwritebuffer
	rc = count;

	return rc;	// return #units written
}


// updates the dwLBAPos counter and dwBytePos counter
unsigned long
cDemuxer::bs_update_lba( Bit_stream_struc *pbs )
{
	// count the #LBAs that we've passed, since the last time this function
	// was called
	while( pbs->dwBytePos >= DEMUX_LBA_SIZE )	// v0.16B34
//	while( pbs->dwBytePos > DEMUX_LBA_SIZE )
	{
		pbs->dwBytePos -= DEMUX_LBA_SIZE;
		pbs->dwLBAPos += 1;
	}

	return pbs->dwBytePos;
}


// 0.16B24 switched to _lseeki64() for >2GB files
long 
cDemuxer::bs_fseek_lba( Bit_stream_struc *pbs, DWORD dwLBA )
{
  __int64 qwPosition = 0;	// target Position
//  long offset = dwLBA * DEMUX_LBA_SIZE;
  __int64 qwOffset = (__int64) dwLBA * DEMUX_LBA_SIZE;
  LPDWORD lpdwordl, lpdwordh;

  if ( !BSOPEN( pbs ) )
  {
	sprintf( m_szText, "bs_fseek_lba( dwLBA = %u ) : No bistream is open!",
		dwLBA );
	TextLogErr( BSERR_FAIL, m_szText );
	return BSERR_FAIL;
  }

  if ( dwLBA < 0 || qwOffset >= pbs->qwLength )
  {
	lpdwordl = (DWORD *) &pbs->qwLength;
	lpdwordh = (DWORD *) &pbs->qwLength + 1;
	sprintf( m_szText, "bs_fseek_lba() : requested LBA (%ld) > bitstream_length ( 0x%X, 0x%X )!",
		dwLBA , *lpdwordh, *lpdwordl );
	TextLogErr( BSERR_FAIL, m_szText );
	return BSERR_FAIL;
  }

  //__int64 _lseeki64( int handle, __int64 offset, int origin );
  //i = fseek( pbs->pt, offset, SEEK_SET );

  bs_update_lba( pbs );	// update LBA variables, v0.16B34 optimization
  // if we are seeking to the NEXT lba, we don't have to use _lseeki64(),
  // we can simply read xxx bytes until we reach the start of the next LBA

  // if user picks '-@ 1', this loop is executed without bs_getbytes()
  // ever being called, causing the program to crash.  Hence the special
  // precaution (dwLBA != 1 ) in the following if statement.
  if ( pbs->dwLBAPos == (dwLBA - 1) && (dwLBA != 1) )
  {
	  bs_getbytes( pbs, m_pBuffer, (DEMUX_LBA_SIZE - pbs->dwBytePos), NULL );
	  bs_update_lba( pbs );
//	  printf("\npbs->dwBytePos = %u, pbs->dwLBAPos = %u", pbs->dwBytePos,
//		pbs->dwLBAPos  );

	  // now check our position, if we didn't get to where we should be,
	  // then the if statement will fall through, and the function will
	  // continue to execute using the my_lseeki64() call.
	  if ( pbs->dwBytePos == 1 && pbs->dwLBAPos == dwLBA )
	  {
//		  printf("\nshortcut seek worked!" );
		  return BSERR_OK;
	  }
  }

  qwPosition = my_lseeki64( pbs->fhandle, qwOffset, SEEK_SET );
  if ( qwPosition == -1 )
  {
	TextLogErr( BSERR_FAIL, "bs_fseek_lba() : my_lseeki64() failed!" );
	return BSERR_FAIL;
  }

  if ( qwOffset != qwPosition )
  {
	sprintf( m_szText, "\nbs_fseek_lba(%ld) : my_lseeki64() failed!", dwLBA );
	TextLogErr( BSERR_FAIL, m_szText );
	return BSERR_FAIL;
  }
/*
  lpdwordl = (LPDWORD) (&qwOffset );
  lpdwordh = (LPDWORD) (&qwOffset + 4);
  printf("\nqwOffset=0x%X 0x%X", *lpdwordl, *lpdwordl );

  lpdwordl = (LPDWORD) (&qwPosition );
  lpdwordh = (LPDWORD) (&qwPosition + 4);
  printf(" qwPosition=0x%X 0x%X", *lpdwordl, *lpdwordl );

  lpdwordl = (LPDWORD) (&qwPosRb );
  lpdwordh = (LPDWORD) (&qwPosRb + 4);
  printf(" qwPosRb=0x%X 0x%X", *lpdwordl, *lpdwordl );
*/
  bs_reset_buffer( pbs );
  //bs_refill_buffer( pbs );  // can't refill here for some reason!
  pbs->dwLBAPos = dwLBA;
  bs_update_lba( pbs );

  return BSERR_OK;
}

DWORD
cDemuxer::bs_get_lba( Bit_stream_struc *pbs, DWORD *lpdword )
{
	if ( !BSOPEN( pbs ) )
	{
		TextLogErr( BSERR_FAIL, "bs_get_lba() : No bitstream is open!");
		return BSERR_FAIL;
	}

	if ( lpdword == NULL )
	{
		TextLogErr( BSERR_FAIL, "bs_get_lba() : NULL pointer!");
		return BSERR_FAIL;
	}

	bs_update_lba( pbs );
	(*lpdword) = pbs->dwLBAPos;
	return BSERR_OK;
}



// returns BSERR_OK if >0 bytes are read (pcbRead is set)
// returns BSERR_FAIL if no bybtes were read (pcbRead not set)

// writes data into the supplied pBuffer[]
// v0.16B34 if ( pcbStored == NULL ), simply skip past data ( N bytes )
// N = # bytes fo read
DWORD  // 0.16B22
cDemuxer::bs_getbytes( 
Bit_stream_struc *pbs, 
unsigned char pBuffer[],  /*byte buffer to store data */
DWORD N, DWORD *pcbStored) /* number of bytes stored in pBuffer[] */
{
	//unsigned char m_pgetbytebuffer[ DEMUX_BUFFER_SIZE ];
	DWORD i = 0;

	if ( N > DEMUX_BUFFER_SIZE )
	{
		fprintf( stderr, "bs_getbytes() : N(%d) exceeds DEMUX_BUFFER_SIZE(%d)",
			N, DEMUX_BUFFER_SIZE );
		return BSERR_FAIL;
	}

	if ( pBuffer == NULL )
	{
		fprintf( stderr, "bs_getbytes() : pcbRead[] == NULL!" );
		return BSERR_FAIL;
	}

	// if at end of current byte, advance to start of next byte
    if (  pbs->buf_bit_idx == 0 )  
	{
      pbs->buf_bit_idx = 8;
      pbs->buf_byte_idx--;
	  ++pbs->dwBytePos;
  	  bs_update_lba( pbs );
	}

	if ( N > pbs->buf_byte_idx - 1 )
	{	// if the buffer is too small, manually order a refill operation
		sprintf( m_szText, "bs_getbytes(%d) : buffer(%d) too small.", N,
			pbs->buf_byte_idx );
		TextLogMsg( LOG_NEW_ENTRY, m_szText );
		printf( "%s", m_szText );
		bs_refill_buffer(pbs );	// v0.16B3 partial fix?
	}

	if ( pbs->buf_bit_idx == 8 )	// are we at a byte-boundary?
	{	// yes, the bitstream-index is byte-aligned, we can just
		// read the bytes without needing to bit-align them

//	  printf("\nbs_getbytes() : pbs->buf_bit_idx == 8!" );
      if ((pbs->buf_byte_idx < DEMUX_MINIMUM) || (pbs->buf_byte_idx < pbs->eob)
		  || ( N > pbs->buf_byte_idx - 1) )
	  {
        if (pbs->eob)
          pbs->eobs = true;	// end of bitstream
        else
        {
          for (i = pbs->buf_byte_idx; i >= 0; i--)
            pbs->buf[pbs->buf_size - 1 - pbs->buf_byte_idx+i] = pbs->buf[i];
          bs_refill_buffer(pbs);
          pbs->buf_byte_idx = pbs->buf_size - 1;
        }
      }

/*	  if ( N > pbs->buf_byte_idx - 1 )
	  {
		  fprintf( stderr, "\nbs_getbytes(%d) : buffer(%d) too small.", N,
			  pbs->buf_byte_idx );
		  // fall through the to the old-fashioned way
	  }
	  else*/
	  {
//		printf("\nbs_getbytes(%ld) : speed-read from buffer(%d)", N,
//		  pbs->buf_byte_idx );

		// copy data directly out of stream buffer.  This is faster
		// than repeatedly calling bs_getbits( 8)
		if ( pcbStored != NULL )
		{
		  for( i = 0; i < N; ++i)
		  {
			m_pBuffer[ i ] = pbs->buf[pbs->buf_byte_idx];
			--pbs->buf_byte_idx;
		  }
		  (*pcbStored ) = N;	// set # bytes actually written to
		}
		else
		{
		  pbs->buf_byte_idx -= N; // advance counter
		}

		pbs->dwBytePos += N;	// advance counter
		bs_update_lba( pbs );

		return BSERR_OK;
	  }	// end if ( pbs->buf_byte_idx < N )
	} //	end if ( pbs->buf_bit_idx == 8 )	// are we at a byte-boundary?

//	printf("\nbs_getbytes(%ld) : using bs_getbits()", N );

	// read each byte as 8bits, "the old fashioned way", slower than
	// doing a direct buffer-copy (above)
	for ( i = 0; i < N; ++i )
		m_pBuffer[ i ] = bs_getbits( pbs, 8 ); // get 8 bits

	if ( pcbStored != NULL )
	  (*pcbStored ) = N;

	pbs->dwBytePos += N;	// advance counter
 	bs_update_lba( pbs );


	return BSERR_OK;
}


// v0.16B24 return the handle to the currently open bitstream
int
cDemuxer::get_fhandle( void) 
{
	return m_bs.fhandle;
}
